#include <iostream>
#include <regex>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <errno.h>
#include <openssl/ssl.h>

using namespace std;

extern int errno;

// This system demonstrates the use of the socket() message passing
// IPC mechanism.  Sockets are different from pipes and shared memory 
// because they can operate BETWEEN computers using the network.
//

int main(void)
{
  ////////////////////////////////////////////////////////////////////
  // Create a socket.  
  ////////////////////////////////////////////////////////////////////  
  int s = socket(AF_INET,SOCK_STREAM,0);

  if (s <= 0) {
    cerr << "clientr: Socket creation failed." << endl;
    exit(-1);
  }

  ////////////////////////////////////////////////////////////////////
  // Connect to server port
  ////////////////////////////////////////////////////////////////////
  sockaddr_in address;
  int addressSize = sizeof(sockaddr_in);
  address.sin_family=AF_INET;                         
  address.sin_port  = htons(443); // <-------------------  port 443 is https
  inet_pton(AF_INET,"",&address.sin_addr);  // <- IP of REST server

  int status=0;
  status = connect(s,(struct sockaddr*) &address, addressSize);

  if (status != 0) {
    cerr << "client:Connect failed, can't connect to port." << endl;
    cout << strerror(errno) << endl;
    exit(-1);
  }

  cout << "Server and client have established communication." << endl;

  //
  // This local buffer will be used to copy messages to/from socket.
  // Sockets require a contiguous block of memory to copy messages.
  // This could be an array of characters, integers, objects, whatever.
  //
  const int BUFFER_SIZE = 50000;
  char buffer[BUFFER_SIZE];

  //
  // first connect to the remote as usual, but use the port 443 instead of 80
  // initialize OpenSSL - do this once and stash ssl_ctx in a global var
  // the flavor of ssl_ctx depends on what version of https is running on
  // the server.  You may need to change the client method.  One way to
  // debug this is to try a simple get with the server using this command
  // line command:
  //    curl -v <url of server plus get url arguments>
  //
  //    curl -v https://api.darksky.net/forecast/543ae8d2205f651fde95f60f5adeb838/40.6884,-75.2207
  // 
  SSL_load_error_strings();
  OpenSSL_add_all_algorithms();
  SSL_library_init ();
  SSL_CTX *ssl_ctx = SSL_CTX_new (TLSv1_client_method());

  // create an SSL connection and attach it to the socket
  SSL *conn = SSL_new(ssl_ctx);
  SSL_set_fd(conn, s);

  // perform the SSL/TLS handshake with the server - when on the
  // server side, this would use SSL_accept()
  int err = SSL_connect(conn);
  if (err <=0)
    {
      int errcode = SSL_get_error(conn, err);
      switch(errcode)
	{
	case SSL_ERROR_NONE: break;        // Cannot happen if err <=0
	case SSL_ERROR_ZERO_RETURN: fprintf(stderr,"SSL connect returned 0.");break;
	case SSL_ERROR_WANT_READ: fprintf(stderr,"SSL connect: Read Error.");break;
	case SSL_ERROR_WANT_WRITE: fprintf(stderr,"SSL connect: Write Error.");break;
	case SSL_ERROR_WANT_CONNECT: fprintf(stderr,"SSL connect: Error connect."); break;
	case SSL_ERROR_WANT_ACCEPT: fprintf(stderr,"SSL connect: Error accept."); break;
	case SSL_ERROR_WANT_X509_LOOKUP: fprintf(stderr,"SSL connect error: X509 lookup."); break;
	case SSL_ERROR_SYSCALL: fprintf(stderr,"SSL connect: Error in system call."); break;
	case SSL_ERROR_SSL: fprintf(stderr,"SSL connect: Protocol Error."); 	  break;
	default: fprintf(stderr,"Failed SSL connect.");
	}
      abort();
    }

  //
  // now proceed with HTTP traffic, using SSL_read instead of recv() and
  // SSL_write instead of send(), and SSL_shutdown/SSL_free before close()
  // use curl to get the format of the expected get request
  //
  cout << "client: Sending get request to server" << endl;
  char* request = "GET /forecast/543ae8d2205f651fde95f60f5adeb838/40.6884,-75.2207 HTTP/1.1\nHost: api.darksky.net\nAccept: */*\n\n";
  SSL_write(conn,request,strlen(request)+1);
  cout << "client: Waiting for server message." << endl;
  int bytesReceived = SSL_read(conn,buffer,BUFFER_SIZE); 

  // regex to process response and get the length of the body of the message
  // totalBytes = message header + message body length - bytes received so far
  regex regexContentLength("Content-Length:\\s\\s*([^\n\r]*)");
  smatch match;
  string str;
  str.assign(buffer);
  regex_search (str,match,regexContentLength);
  int totalBytes = stoi(match[1]) - bytesReceived;

  // regex to find position of start of body of message
  // this is the size of the message header add this to totalBytes
  regex regexBody("\\r\\n\\r\\n");
  regex_search(str,match,regexBody);
  totalBytes += match.position(0)+4;

  // read the rest of the message body
  while (totalBytes > 0)
  {
    bzero(buffer,BUFFER_SIZE);
    totalBytes -= SSL_read(conn,buffer,totalBytes);
    str.append(buffer);
  }

  cout << str << endl;
  
  cout.flush();
  close(s);
  
  return 1;
}
